import json

from phy_credit.imageV2 import (
    CategorizationDef,
    LabelInterface,
    build_label_info,
)


def test_build_label_info_from_scratch(
    label_interface_sample,
    add_object_1,
    add_object_2,
    add_object_3,
    add_object_4,
    sample_tag_result,
    sample_info_result,
):
    label_interface = LabelInterface.from_dict(label_interface_sample)
    labelInfo = build_label_info(label_interface)
    labelInfo.add_object(**add_object_1)
    labelInfo.add_object(**add_object_2)
    labelInfo.add_object(**add_object_3)
    labelInfo.add_object(**add_object_4)

    tags = labelInfo.build_tags()
    assert json.dumps(tags, sort_keys=True) == json.dumps(
        sample_tag_result, sort_keys=True
    )
    # TODO: Comparison is a bit complex because of IDs
    # info = labelInfo.build_info()
    # assert json.dumps(info, sort_keys=True) == json.dumps(
    #     sample_info_result, sort_keys=True)


def test_build_label_info_from_result(
    label_interface_sample,
    sample_build_result,
    sample_tag_result,
    sample_info_result,
):
    label_interface = LabelInterface.from_dict(label_interface_sample)
    labelInfo = build_label_info(
        label_interface=label_interface,
        result=sample_build_result,
    )

    tags = labelInfo.build_tags()
    assert json.dumps(tags, sort_keys=True) == json.dumps(
        sample_tag_result, sort_keys=True
    )
    info = labelInfo.build_info()
    assert json.dumps(info, sort_keys=True) == json.dumps(
        sample_info_result, sort_keys=True
    )


def test_build_label_info_from_scratch_v4(
    label_interface_sample_v4,
    add_object_1,
    add_object_2,
    set_category_1,
    sample_tag_result_v4,
    sample_info_result_v4,
):

    label_interface = LabelInterface.from_dict(label_interface_sample_v4)
    label_info = build_label_info(label_interface)
    label_info.add_object(**add_object_1)
    label_info.add_object(**add_object_2)
    categorization = CategorizationDef.from_dict(set_category_1)
    label_info.set_categories(categorization)

    tags = label_info.build_tags()
    assert json.dumps(tags, sort_keys=True) == json.dumps(
        sample_tag_result_v4, sort_keys=True
    )
    assert len(tags["classes_properties_count"]) == 4
    # # TODO: Comparison is a bit complex because of IDs
    # info = label_info.build_info()
    # assert json.dumps(info, sort_keys=True) == json.dumps(
    #     sample_info_result_v4, sort_keys=True
    # )


# using creators
def test_build_label_info_from_scratch_v6(
    label_interface_sample_v6, label_info_sample_v6
):
    label_interface = LabelInterface.from_dict(label_interface_sample_v6)
    label_info = build_label_info(label_interface)
    box_creator = label_info.get_box_creator()

    box = box_creator.create(
        id="12590dbe-72f1-4278-b1fe-b4a0642d55b6",
        class_name="person",
        coord={"x": 26.3, "y": 24.1, "width": 36.3, "height": 25.2},
        meta={
            "alpha": 1,
            "color": "#FF625A",
            "visible": True,
            "z_index": 1
        }
    )

    label_info.add_object(**box.to_dict())

    rbox_creator = label_info.get_rotated_box_creator()

    rbox = rbox_creator.create(
        id="6f4e5ef6-e3ea-4a2d-b0a1-3ebd00b965b3",
        class_name="person",
        coord={
            "cx": 55.6,
            "cy": 84.0,
            "width": 27.8,
            "height": 22.4,
            "angle": 24.3,
        },
        meta={
            "alpha": 1,
            "color": "#FE9573",
            "visible": True,
            "z_index": 1
        }
    )

    label_info.add_object(**rbox.to_dict())

    polyline_creator = label_info.get_polyline_creator()

    polyline = polyline_creator.create(
        id="9b20fabd-9b33-422b-bbb5-870cfaad31b2",
        class_name="poly",
        coord={
            "points": [
                [
                    {"x": 20.5, "y": 144.8},
                    {"x": 40.9, "y": 117.3},
                    {"x": 40.8, "y": 145.8},
                ]
            ]
        },
        meta={
            "alpha": 1,
            "color": "#FFAF5A",
            "visible": True,
            "z_index": 1
        }
    )

    label_info.add_object(**polyline.to_dict())

    polygon_creator = label_info.get_polygon_creator()

    polygon = polygon_creator.create(
        id="92e28cc4-7eee-4a72-9dc4-383ffd4ecf70",
        class_name="poly_seg",
        coord={
            "points": [
                [
                    [
                        {"x": 21.5, "y": 191.2},
                        {"x": 24.2, "y": 167.7},
                        {"x": 28.1, "y": 158.1},
                        {"x": 41.2, "y": 161.9},
                        {"x": 21.5, "y": 191.2},
                    ],
                ]
            ]
        },
        meta={
            "alpha": 1,
            "color": "#FFCC00",
            "visible": True,
            "z_index": 1
        }
    )

    label_info.add_object(**polygon.to_dict())

    keypoint_cretor = label_info.get_keypoint_creator()
    default_kp_coord = keypoint_cretor.get_default_keypoint_coord(
        keypoint_interface_id="facial-landmark-15"
    )
    keypoint = keypoint_cretor.create(
        id="9ed93108-2c3a-4eba-8613-0633003a239f",
        class_name="face_kp",
        coord=default_kp_coord,
        meta={
            "alpha": 1,
            "color": "#FFF73E",
            "visible": True,
            "z_index": 1
        }
    )

    label_info.add_object(**keypoint.to_dict())

    cuboid2d_creator = label_info.get_2D_cubid_creator()

    cuboid_2d = cuboid2d_creator.create(
        id="10a57f9d-3d04-4ce6-aa8f-516300cd8f75",
        class_name="2d_cuboid",
        coord={
            "near": {"x": 15.1, "y": 225.7, "width": 44.9, "height": 37.6},
            "far": {"x": 83.2, "y": 239.3, "width": 5.4, "height": 16.5},
        },
        meta={
            "alpha": 1,
            "color": "#DEF00F",
            "visible": True,
            "z_index": 1
        }
    )
    label_info.add_object(**cuboid_2d.to_dict())

    assert json.dumps(label_info.build_info(), sort_keys=True) == json.dumps(
        label_info_sample_v6, sort_keys=True
    )



from typing import Any, Dict, List, Union
from .label import ClassType

JsonDict = Dict[str, Any]


def add_class_id(property: JsonDict, object: JsonDict):
    return {**property, "class_id": object["class_id"]}


def count_properties(properties):
    result = {}

    for property in properties:
        unique_option_id = property["property_option_id"]
        if unique_option_id not in result:
            result[unique_option_id] = {**property, "count": 1}
        else:
            result[unique_option_id]["count"] += 1

    return list(result.values())


def _unique_option_id(property_id: str, option_id: str) -> str:
    return f"{property_id}#{option_id}"


def _format_radio_option(property) -> JsonDict:
    return {
        "class_id": property["class_id"],
        "property_id": property["property_id"],
        "property_name": property["property_name"],
        "option_id": property["option_id"],
        "option_name": property["option_name"],
        "property_option_id": _unique_option_id(
            property["property_id"], property["option_id"]
        ),
    }


def _flatten_and_format_checkbox_options(property: JsonDict) -> List[JsonDict]:
    option_ids = property["option_ids"]
    option_names = property["option_names"]

    return [
        {
            "class_id": property["class_id"],
            "property_id": property["property_id"],
            "property_name": property["property_name"],
            "option_id": option_id,
            "option_name": option_names[idx],
            "property_option_id": _unique_option_id(
                property["property_id"], option_id
            ),
        }
        for (idx, option_id) in enumerate(option_ids)
    ]


def expand_options(properties: List[JsonDict]) -> List[JsonDict]:
    result = []

    def check_valid_radio_option(property):
        return property.get("option_id") is not None

    def check_valid_checkbox(property):
        # check fields not null and has len > 0
        return property["option_ids"] and property["option_names"]

    def list_options(property):
        if property.get("type") == "radio" and check_valid_radio_option(
            property
        ):
            return [_format_radio_option(property)]
        if property.get("type") == "checkbox" and check_valid_checkbox(
            property
        ):
            return _flatten_and_format_checkbox_options(property)
        return []

    return [o for property in properties for o in list_options(property)]


def remove_free_response(properties: List[JsonDict]) -> List[JsonDict]:
    return [
        property
        for property in properties
        if property.get("type") in ["radio", "checkbox"]
    ]


def get_opt_map(options: list) -> dict:
    res = {}
    for opt in options:
        if "children" in opt:
            res.update(get_opt_map(opt["children"]))
        else:
            res[opt["name"]] = opt
    return res


def set_properties(properties_def, properties):
    prop_def_map = {prop_def["name"]: prop_def for prop_def in properties_def}
    converted_properties = []
    if properties and isinstance(properties, list):
        for prop in properties:
            prop_def = prop_def_map[prop["name"]]
            if prop_def["type"] in ["radio", "dropdown", "checkbox"]:
                opt_map = get_opt_map(prop_def["options"])
                if prop_def["type"] == "checkbox":
                    converted_properties.append(
                        {
                            "type": prop_def["type"],
                            "property_id": prop_def["id"],
                            "property_name": prop_def["name"],
                            "option_ids": [
                                opt_map[val]["id"] for val in prop["value"]
                            ],
                            "option_names": [
                                opt_map[val]["name"] for val in prop["value"]
                            ],
                        }
                    )
                else:
                    converted_properties.append(
                        {
                            "type": prop_def["type"],
                            "property_id": prop_def["id"],
                            "property_name": prop_def["name"],
                            "option_id": opt_map[prop["value"]]["id"],
                            "option_name": opt_map[prop["value"]]["name"],
                        }
                    )
            elif prop_def["type"] == "free response":
                converted_properties.append(
                    {
                        "type": prop_def["type"],
                        "property_id": prop_def["id"],
                        "property_name": prop_def["name"],
                        "value": prop["value"],
                    }
                )
    return converted_properties


def get_properties(properties_def, properties):
    prop_def_map = {prop_def["name"]: prop_def for prop_def in properties_def}
    converted_properties = []
    for prop in properties:
        prop_def = prop_def_map[prop["property_name"]]
        if prop_def["type"] in ["radio", "dropdown"]:
            converted_properties.append(
                {
                    "name": prop["property_name"],
                    "value": prop["option_name"],
                }
            )
        elif prop_def["type"] == "checkbox":
            converted_properties.append(
                {
                    "name": prop["property_name"],
                    "value": prop["option_names"],
                }
            )
        elif prop_def["type"] == "free response":
            converted_properties.append(
                {
                    "name": prop["property_name"],
                    "value": prop["value"],
                }
            )
    return converted_properties


def dim(arr):
    if not type(arr) == list:
        return []
    return [len(arr)] + dim(arr[0])


def unique_string_builder(name: str, type: Union[str, 'ClassType']):
    assert ClassType.is_valid_class_type(type), f"Invalid class type: {type}"
    if isinstance(type, ClassType):
        return f"{name}##_##{type.value}"
    elif isinstance(type, str):
        return f"{name}##_##{type}"
    else:
        raise TypeError(f"Invalid type: {type}")


def is_sublist(a: list, b: list) -> bool:
    for i in a:
        if i not in b:
            return False
    return True



from collections import Counter
from typing import Any, Dict, List

from ..common.utils import (add_class_id, count_properties,
                            expand_options, remove_free_response)

JsonDict = Dict[str, Any]


def get_video_frame_properties(object: JsonDict) -> List[JsonDict]:
    result = []
    for frame in object.get("frames", []):
        properties_in_frame = [
            add_class_id(property, object) for property in frame.get("properties", [])
        ]
        result.extend(properties_in_frame)
    return result


def get_video_label_properties(object: JsonDict) -> List[JsonDict]:
    return [add_class_id(property, object) for property in object.get("properties", [])]


def flatten_video_obj_properties(objects: List[JsonDict]) -> List[JsonDict]:
    result = []
    for object in objects:
        result.extend(get_video_frame_properties(object))
        result.extend(get_video_label_properties(object))
    return result


def preprocess_video_obj_properties(objects: List[JsonDict]) -> List[JsonDict]:
    return expand_options(remove_free_response(flatten_video_obj_properties(objects)))


def calculate_video_properties_count(objects):
    properties = preprocess_video_obj_properties(objects)
    return count_properties(properties)


def calculate_annotated_frame_count(objects: List[JsonDict]) -> int:
    # include frames which have annotated objects (exclude categories)
    annotations = [
        annotation for obj in objects for annotation in obj["frames"]]
    frame_number_to_anno_count = Counter(anno["num"] for anno in annotations)
    return len(frame_number_to_anno_count)



from ..common import is_camel_version
from .categorization import CategorizationDef
from .label_info import LabelInfo
from .label_interface import LabelInterface
from .legacy.label_info import LabelInfo as LegacyLabelInfo
from .object_tracking import ObjectTrackingDef
from .project_info import ProjectInfo
from .property import PropertyDef, PropertyOptionsDef, PropertyOptionsItemDef


def build_label_info(label_interface, result=None):
    if type(label_interface) is dict:
        label_interface = LabelInterface.from_dict(label_interface)
    if is_camel_version(label_interface):
        return LegacyLabelInfo(label_interface, result)
    return LabelInfo(label_interface, result)


__all__ = (
    "LabelInfo",
    "LabelInterface",
    "ObjectTrackingDef",
    "LegacyLabelInfo",
    "build_label_info",
    "ProjectInfo",
    "CategorizationDef",
    "PropertyDef",
    "PropertyOptionsDef",
    "PropertyOptionsItemDef",
)



from ..video.categorization import CategorizationDef
# from .label_info import LabelInfo
from .label_interface import LabelInterface
from .object_tracking import ObjectTrackingDef
from .project_info import ProjectInfo
from ..video.property import PropertyDef, PropertyOptionsDef, PropertyOptionsItemDef


# def build_label_info(label_interface, result=None):
#     if type(label_interface) is dict:
#         label_interface = LabelInterface.from_dict(label_interface)
#     return LabelInfo(label_interface, result)


__all__ = (
    # "LabelInfo",
    "LabelInterface",
    "ObjectTrackingDef",
    # "build_label_info",
    "ProjectInfo",
    "CategorizationDef",
    "PropertyDef",
    "PropertyOptionsDef",
    "PropertyOptionsItemDef",
)



from abc import ABC
from enum import Enum
from typing import Union


class ClassType(Enum):
    BOX = "box"
    RBOX = "rbox"
    POLYLINE = "polyline"
    POLYGON = "polygon"
    KEYPOINT = "keypoint"
    CUBOID2D = "cuboid2d"
    CUBOID = "cuboid"

    @staticmethod
    def is_valid_class_type(class_type: Union[str, 'ClassType']):
        if isinstance(class_type, ClassType):
            return True
        elif isinstance(class_type, str):
            return class_type in [member.value for member in ClassType]
        else:
            return False


class Label(ABC):
    def __init__(self, attributes):
        self.id = attributes["id"]
        self.class_name = attributes["class_name"]
        self.class_id = attributes["class_id"]
        self.properties_def = attributes["properties_def"]
        self.properties = attributes["properties"]
        self.tracking_id = attributes["tracking_id"]
        self.annotation_type = None

    def to_dict(self):
        pass



import json

from phy_credit.imageV2 import (
    CategorizationDef,
    LabelInterface,
    ObjectDetectionDef,
    PropertyDef,
    PropertyOptionsDef,
    PropertyOptionsItemDef,
)


def test_build_label_interface_v6_from_scratch(label_interface_sample_v6):
    categorization = CategorizationDef.get_default()
    ms_1 = PropertyOptionsItemDef(
        name="ms_1", id="14d9bb96-17cf-44d0-9929-4a9144d85797"
    )
    ms_2 = PropertyOptionsItemDef(
        name="ms_2", id="eba4b04b-b23f-4e65-a020-88e6731300f4"
    )
    ms_g = PropertyOptionsDef(
        name="ms_g",
        id="13cb5be0-33cc-4dda-84c3-ec6afb50680c",
        children=[ms_1, ms_2],
    )
    categorization.add_multiple_selection_category(
        name="ms_cate",
        id="bb0e4235-9d17-40ed-9852-26fc66941d05",
        options=[ms_g],
        is_required=False,
        description="",
        render_value=False,
        default_value=[],
    )
    mc_1 = PropertyOptionsItemDef(
        name="mc_1", id="8831d4a4-5bee-4252-b651-c254b4593490"
    )
    mc_2 = PropertyOptionsItemDef(
        name="mc_2", id="08e881d0-c072-443b-9887-d4042c09a998"
    )
    mc_g = PropertyOptionsDef(
        name="mc_g",
        id="0bd3a095-2e61-4cf3-b7aa-cb58d4ef1912",
        children=[mc_1, mc_2],
    )
    categorization.add_multiple_choice_category(
        name="mc_cate",
        id="0da7b178-df30-47da-bf9b-2daadfbc99d7",
        options=[mc_g],
        is_required=False,
        description="",
        render_value=False,
        default_value="",
    )
    categorization.add_free_response_category(
        name="fr_cate",
        id="69f0963a-0400-4d9a-81fe-b7ef3ded55db",
        blank=True,
        description="",
        render_value=False,
        default_value="",
    )
    assert json.dumps(categorization.to_dict(), sort_keys=True) == json.dumps(
        label_interface_sample_v6["categorization"], sort_keys=True
    )
    object_detection = ObjectDetectionDef.get_default()
    opt_1 = PropertyOptionsItemDef(name="Untitled Option", id="1")
    opt_2 = PropertyOptionsItemDef(name="Untitled Option (1)", id="2")
    box_prop = PropertyDef.multiple_choice_property(
        name="box Property",
        options=[opt_1, opt_2],
        id="c126812b-c013-4f78-b709-7a1764c7c7f3",
        default_value=None,
        is_required=True,
    )
    opt_3 = PropertyOptionsItemDef(name="Untitled Option", id="3")
    opt_4 = PropertyOptionsItemDef(name="Untitled Option (1)", id="4")
    box_prop_1 = PropertyDef.multiple_selection_property(
        name="box Property (1)",
        options=[opt_3, opt_4],
        id="3b09f755-e34f-4d5b-9950-0467d08b5c6b",
        default_value=[],
        is_required=True,
    )
    box_prop_2 = PropertyDef.free_response_property(
        name="box Property (2)",
        id="da2e15ff-98bc-4ef4-a1c9-1830f0e32908",
        blank=False,
        description="",
        render_value=False,
        default_value="",
        constraints={
            "alphabet": True,
            "digit": True,
            "space": True,
            "special": True,
        },
    )
    object_detection.add_box(
        name="person",
        id="346c2e59-0d4d-4818-8cd0-538b157fa06c",
        color="#FF625A",
        properties=[box_prop, box_prop_1, box_prop_2],
    )
    assert json.dumps(
        object_detection.to_dict()["object_classes"][0], sort_keys=True
    ) == json.dumps(
        label_interface_sample_v6["object_detection"]["object_classes"][0],
        sort_keys=True,
    )
    rbox_prop = PropertyDef.multiple_choice_property(
        name="rbox Property",
        options=[opt_1, opt_2],
        id="e8405ff2-a091-4fb3-9ee3-33921491be6b",
        is_required=True,
    )
    rbox_prop_1 = PropertyDef.multiple_selection_property(
        name="rbox Property (1)",
        options=[opt_3, opt_4],
        id="3e75a099-875a-4576-9548-2e2f171e6ad5",
        is_required=True,
    )
    rbox_prop_2 = PropertyDef.free_response_property(
        name="rbox Property (2)",
        id="29243fd2-92c4-4f11-baa0-0e615901039f",
        blank=False,
        description="",
        render_value=False,
        default_value="",
        constraints={
            "alphabet": True,
            "digit": True,
            "space": True,
            "special": True,
        },
    )
    object_detection.add_rbox(
        name="person",
        id="e15fae76-1702-44e9-97d8-25bf420e4dbe",
        color="#FE9573",
        properties=[rbox_prop, rbox_prop_1, rbox_prop_2],
    )
    assert json.dumps(
        object_detection.to_dict()["object_classes"][1], sort_keys=True
    ) == json.dumps(
        label_interface_sample_v6["object_detection"]["object_classes"][1],
        sort_keys=True,
    )
    poly_prop = PropertyDef.multiple_choice_property(
        name="poly Property",
        options=[opt_1, opt_2],
        id="d1516ef1-a64d-45c4-9bd0-d26a38658839",
        is_required=True,
    )
    poly_prop_1 = PropertyDef.multiple_selection_property(
        name="poly Property (1)",
        options=[opt_3, opt_4],
        id="7ac2d2d6-bbcf-422e-8b88-0b20335626a0",
        is_required=True,
    )
    poly_prop_2 = PropertyDef.free_response_property(
        name="poly Property (2)",
        id="a2bae19d-b700-4633-81b3-880b6fcc9fc0",
        blank=False,
        description="",
        render_value=False,
        default_value="",
        constraints={
            "alphabet": True,
            "digit": True,
            "space": True,
            "special": True,
        },
    )
    object_detection.add_polyline(
        name="poly",
        id="7c6ab4db-5b0c-4143-b992-6dc732f28870",
        color="#FFAF5A",
        properties=[poly_prop, poly_prop_1, poly_prop_2],
    )
    assert json.dumps(
        object_detection.to_dict()["object_classes"][2], sort_keys=True
    ) == json.dumps(
        label_interface_sample_v6["object_detection"]["object_classes"][2],
        sort_keys=True,
    )
    poly_seg_prop = PropertyDef.multiple_choice_property(
        name="poly_seg Property",
        options=[opt_1, opt_2],
        id="00f9510e-c574-4667-8f76-17c51618e272",
        is_required=True,
    )
    poly_seg_prop_1 = PropertyDef.multiple_selection_property(
        name="poly_seg Property (1)",
        options=[opt_3, opt_4],
        id="1007f627-a13f-4678-a411-871482dfd166",
        is_required=True,
    )
    poly_seg_prop_2 = PropertyDef.free_response_property(
        name="poly_seg Property (2)",
        id="2fccbadc-271f-4124-a3e2-71a759db27c0",
        blank=False,
        description="",
        render_value=False,
        default_value="",
        constraints={
            "alphabet": True,
            "digit": True,
            "space": True,
            "special": True,
        },
    )
    object_detection.add_polygon(
        name="poly_seg",
        id="5a122ca5-b827-4896-9476-ab01d5164853",
        color="#FFCC00",
        properties=[poly_seg_prop, poly_seg_prop_1, poly_seg_prop_2],
    )
    assert json.dumps(
        object_detection.to_dict()["object_classes"][3], sort_keys=True
    ) == json.dumps(
        label_interface_sample_v6["object_detection"]["object_classes"][3],
        sort_keys=True,
    )

    # key point interface can be built on suite project settings.
    face_kp_interface = {
        "id": "facial-landmark-15",
        "name": "face",
        "edges": [
            {"u": 1, "v": 0, "color": "#64b5f6"},
            {"u": 2, "v": 0, "color": "#64b5f6"},
            {"u": 4, "v": 3, "color": "#ff8a65"},
            {"u": 5, "v": 3, "color": "#ff8a65"},
            {"u": 7, "v": 6, "color": "#64b5f6"},
            {"u": 9, "v": 8, "color": "#ff8a65"},
            {"u": 13, "v": 11, "color": "#64b5f6"},
            {"u": 14, "v": 11, "color": "#64b5f6"},
            {"u": 13, "v": 12, "color": "#ff8a65"},
            {"u": 14, "v": 12, "color": "#ff8a65"},
        ],
        "points": [
            {
                "name": "left eye center",
                "color": "#64b5f6",
                "default_value": {
                    "x": 0.75,
                    "y": 0.25,
                    "state": {"visible": True},
                },
                "symmetric_idx": 3,
            },
            {
                "name": "left eye inner corner",
                "color": "#64b5f6",
                "default_value": {
                    "x": 0.625,
                    "y": 0.25,
                    "state": {"visible": True},
                },
                "symmetric_idx": 4,
            },
            {
                "name": "left eye outer corner",
                "color": "#64b5f6",
                "default_value": {
                    "x": 0.875,
                    "y": 0.25,
                    "state": {"visible": True},
                },
                "symmetric_idx": 5,
            },
            {
                "name": "right eye center",
                "color": "#ff8a65",
                "default_value": {
                    "x": 0.25,
                    "y": 0.25,
                    "state": {"visible": True},
                },
                "symmetric_idx": 0,
            },
            {
                "name": "right eye inner corner",
                "color": "#ff8a65",
                "default_value": {
                    "x": 0.375,
                    "y": 0.25,
                    "state": {"visible": True},
                },
                "symmetric_idx": 1,
            },
            {
                "name": "right eye outer corner",
                "color": "#ff8a65",
                "default_value": {
                    "x": 0.125,
                    "y": 0.25,
                    "state": {"visible": True},
                },
                "symmetric_idx": 2,
            },
            {
                "name": "left eyebrow inner end",
                "color": "#64b5f6",
                "default_value": {
                    "x": 0.625,
                    "y": 0,
                    "state": {"visible": True},
                },
                "symmetric_idx": 8,
            },
            {
                "name": "left eyebrow outer end",
                "color": "#64b5f6",
                "default_value": {
                    "x": 1,
                    "y": 0,
                    "state": {"visible": True},
                },
                "symmetric_idx": 9,
            },
            {
                "name": "right eyebrow inner end",
                "color": "#ff8a65",
                "default_value": {
                    "x": 0.375,
                    "y": 0,
                    "state": {"visible": True},
                },
                "symmetric_idx": 6,
            },
            {
                "name": "right eyebrow outer end",
                "color": "#ff8a65",
                "default_value": {
                    "x": 0,
                    "y": 0,
                    "state": {"visible": True},
                },
                "symmetric_idx": 7,
            },
            {
                "name": "nose tip",
                "color": "#d50000",
                "default_value": {
                    "x": 0.5,
                    "y": 0.5,
                    "state": {"visible": True},
                },
            },
            {
                "name": "mouth left corner",
                "color": "#64b5f6",
                "default_value": {
                    "x": 0.75,
                    "y": 0.875,
                    "state": {"visible": True},
                },
                "symmetric_idx": 12,
            },
            {
                "name": "mouth right corner",
                "color": "#ff8a65",
                "default_value": {
                    "x": 0.25,
                    "y": 0.875,
                    "state": {"visible": True},
                },
                "symmetric_idx": 11,
            },
            {
                "name": "mouth center top lip",
                "color": "#d50000",
                "default_value": {
                    "x": 0.5,
                    "y": 0.75,
                    "state": {"visible": True},
                },
                "symmetric_idx": 13,
            },
            {
                "name": "mouth center bottom lip",
                "color": "#d50000",
                "default_value": {
                    "x": 0.5,
                    "y": 1,
                    "state": {"visible": True},
                },
                "symmetric_idx": 14,
            },
        ],
        "allow_valid_invisibles": False,
    }
    face_kp_prop = PropertyDef.multiple_choice_property(
        name="face_kp Property",
        options=[opt_1, opt_2],
        id="8f826174-6b13-4144-80a9-df236d986c38",
        is_required=True,
    )
    face_kp_prop_1 = PropertyDef.multiple_selection_property(
        name="face_kp Property (1)",
        options=[opt_3, opt_4],
        id="743e543b-c51c-4eac-a29b-f8fa20dd2474",
        is_required=True,
    )
    face_kp_prop_2 = PropertyDef.free_response_property(
        name="face_kp Property (2)",
        id="931e64e5-30b2-4da0-a435-b32f40cad7cb",
        blank=False,
        description="",
        render_value=False,
        default_value="",
        constraints={
            "alphabet": True,
            "digit": True,
            "space": True,
            "special": True,
        },
    )
    object_detection.add_keypoint(
        name="face_kp",
        id="96007e8f-3c76-4359-88d3-27cc3b2e986f",
        keypoint_id="facial-landmark-15",
        color="#FFF73E",
        properties=[face_kp_prop, face_kp_prop_1, face_kp_prop_2],
        keypoint_template=face_kp_interface,
    )
    assert json.dumps(
        object_detection.to_dict()["keypoints"], sort_keys=True
    ) == json.dumps(
        label_interface_sample_v6["object_detection"]["keypoints"],
        sort_keys=True,
    )
    assert json.dumps(
        object_detection.to_dict()["object_classes"][4], sort_keys=True
    ) == json.dumps(
        label_interface_sample_v6["object_detection"]["object_classes"][4],
        sort_keys=True,
    )
    cuboid_2d_prop = PropertyDef.multiple_choice_property(
        name="2d_cuboid Property",
        options=[opt_1, opt_2],
        id="cd1ce733-f0f9-45f2-b2d3-d95016807656",
        is_required=True,
    )
    cuboid_2d_prop_1 = PropertyDef.multiple_selection_property(
        name="2d_cuboid Property (1)",
        options=[opt_3, opt_4],
        id="faf00cd0-99d6-409e-a9ad-75c8830c021a",
        is_required=True,
    )
    cuboid_2d_prop_2 = PropertyDef.free_response_property(
        name="2d_cuboid Property (2)",
        id="b75b8e1f-0e18-405f-bcde-f808fc3595dc",
        blank=False,
        description="",
        render_value=False,
        default_value="",
        constraints={
            "alphabet": True,
            "digit": True,
            "space": True,
            "special": True,
        },
    )
    object_detection.add_2dcuboid(
        name="2d_cuboid",
        id="e970ec55-2dd7-4a94-8a93-ec2b34ed5005",
        color="#DEF00F",
        properties=[cuboid_2d_prop, cuboid_2d_prop_1, cuboid_2d_prop_2],
    )
    assert json.dumps(
        object_detection.to_dict()["object_classes"][5], sort_keys=True
    ) == json.dumps(
        label_interface_sample_v6["object_detection"]["object_classes"][5],
        sort_keys=True,
    )
    object_detection.add_box(
        name="박스", id="414d1b1d-fe14-4e16-af4e-908a5a8147c1", color="#A3EB57"
    )
    assert json.dumps(
        object_detection.to_dict()["object_classes"][6], sort_keys=True
    ) == json.dumps(
        label_interface_sample_v6["object_detection"]["object_classes"][6],
        sort_keys=True,
    )

    label_interface = LabelInterface.get_default()
    label_interface.set_categorization(categorization=categorization)
    label_interface.set_object_detection(object_detection=object_detection)

    assert json.dumps(
        label_interface.object_detection.to_dict(), sort_keys=True
    ) == json.dumps(
        label_interface_sample_v6["object_detection"], sort_keys=True
    )


def test_build_label_interface_v4_from_scratch(label_interface_sample_v4):
    option_a = PropertyOptionsItemDef(
        name="a", id="b2ad46b3-ab05-41e1-be71-6b454daadf8c"
    )
    option_b = PropertyOptionsItemDef(
        name="b", id="cd0df355-75f3-48c0-977a-8f2b7cd8197b"
    )
    option_c = PropertyOptionsItemDef(
        name="c", id="de5f7920-e9bc-441f-b4c0-3ebd8fe7013a"
    )
    option_alphabet = PropertyOptionsDef(
        name="alphabet",
        id="f2c19b26-37fa-4b74-baaf-1105a4b51841",
        children=[option_a, option_b, option_c],
    )

    option_1 = PropertyOptionsItemDef(
        name="1", id="a1200bca-41c2-43e5-85d9-4d51554d552b"
    )
    option_2 = PropertyOptionsItemDef(
        name="2", id="d6711d55-aae5-4dbe-a92e-55d87c1c02ef"
    )
    option_3 = PropertyOptionsItemDef(
        name="3", id="87c0bf6a-6dfb-46d8-99b4-5112340529c9"
    )
    option_number = PropertyOptionsDef(
        name="number",
        id="4964136d-8ebf-4692-98ee-8f81b0fb1234",
        children=[option_1, option_2, option_3],
    )

    categorization = CategorizationDef.get_default()
    categorization.add_multiple_selection_category(
        name="Root",
        id="root",
        options=[option_alphabet, option_number],
        is_required=True,
    )

    # print(categorization.to_dict())
    object_detection = ObjectDetectionDef.get_default()
    pc_option_1 = PropertyOptionsItemDef(name="1", id="1")
    pc_option_2 = PropertyOptionsItemDef(name="2", id="2")
    pc_properties = PropertyDef.multiple_choice_property(
        name="pc",
        id="5b65d4ae-2737-4de8-a460-70a5aab4eba2",
        options=[pc_option_1, pc_option_2],
        is_required=True,
    )
    ps_option_1 = PropertyOptionsItemDef(name="1", id="3")
    ps_option_2 = PropertyOptionsItemDef(name="2", id="4")
    ps_properties = PropertyDef.multiple_selection_property(
        name="ps",
        id="f7eebc07-155b-48f3-847d-5501668044ad",
        options=[ps_option_1, ps_option_2],
        is_required=True,
    )

    object_detection.add_box(
        name="bb",
        id="6b5ec356-cf5d-4e73-9cfb-1f7a3393f4a9",
        color="#FF625A",
        properties=[pc_properties, ps_properties],
    )

    label_interface = LabelInterface.get_default()

    label_interface.set_categorization(categorization=categorization)
    label_interface.set_object_detection(object_detection=object_detection)

    assert label_interface_sample_v4 == label_interface.to_dict()
    # print(json.dumps(label_interface.to_dict(), indent=4))
    # # This will print the result below


"""
{
    "type": "image-siesta",
    "version": "0.4.7",
    "data_type": "image",
    "categorization": {
        "properties": [
            {
                "id": "root",
                "name": "Root",
                "default_value": [],
                "description": "",
                "options": [
                    {
                        "name": "alphabet",
                        "id": "f2c19b26-37fa-4b74-baaf-1105a4b51841",
                        "children": [
                            {
                                "id": "b2ad46b3-ab05-41e1-be71-6b454daadf8c",
                                "name": "a"
                            },
                            {
                                "id": "cd0df355-75f3-48c0-977a-8f2b7cd8197b",
                                "name": "b"
                            },
                            {
                                "id": "de5f7920-e9bc-441f-b4c0-3ebd8fe7013a",
                                "name": "c"
                            }
                        ]
                    },
                    {
                        "name": "number",
                        "id": "4964136d-8ebf-4692-98ee-8f81b0fb1234",
                        "children": [
                            {
                                "id": "a1200bca-41c2-43e5-85d9-4d51554d552b",
                                "name": "1"
                            },
                            {
                                "id": "d6711d55-aae5-4dbe-a92e-55d87c1c02ef",
                                "name": "2"
                            },
                            {
                                "id": "87c0bf6a-6dfb-46d8-99b4-5112340529c9",
                                "name": "3"
                            }
                        ]
                    }
                ],
                "render_value": false,
                "required": true,
                "type": "checkbox"
            }
        ]
    },
    "object_detection": {
        "keypoints": [],
        "object_groups": [],
        "object_classes": [
            {
                "id": "6b5ec356-cf5d-4e73-9cfb-1f7a3393f4a9",
                "name": "bb",
                "color": "#FF625A",
                "properties": [
                    {
                        "id": "5b65d4ae-2737-4de8-a460-70a5aab4eba2",
                        "name": "pc",
                        "default_value": null,
                        "description": "",
                        "options": [
                            {
                                "id": "1",
                                "name": "1"
                            },
                            {
                                "id": "2",
                                "name": "2"
                            }
                        ],
                        "render_value": false,
                        "required": true,
                        "type": "radio"
                    },
                    {
                        "id": "f7eebc07-155b-48f3-847d-5501668044ad",
                        "name": "ps",
                        "default_value": [],
                        "description": "",
                        "options": [
                            {
                                "id": "3",
                                "name": "1"
                            },
                            {
                                "id": "4",
                                "name": "2"
                            }
                        ],
                        "render_value": false,
                        "required": true,
                        "type": "checkbox"
                    }
                ],
                "constraints": {},
                "ai_class_map": [],
                "annotation_type": "box"
            }
        ],
        "annotation_types": [
            "box",
            "image category"
        ]
    }
}
"""



from uuid import uuid4

from ..common.label_creator import LabelCreator
from ..common.label import ClassType
from ..common.utils import dim, unique_string_builder, is_sublist
from .image_label import Box, Cuboid2D, Keypoint, Polygon, Polyline, RotatedBox


class BoxCreator(LabelCreator):
    def __init__(self, object_classes_map):
        super().__init__(object_classes_map)

    def get_default_box_coord(self):
        return {
            "x": 0.0,
            "y": 0.0,
            "width": 0.0,
            "height": 0.0,
        }

    def create(
        self,
        class_name: str,
        coord: dict,
        meta: dict = {},
        properties: list = [],
        id: str = str(uuid4()),
        tracking_id: int = None,
    ):
        assert unique_string_builder(
            class_name,
            ClassType.BOX,
        ) in self.object_classes_map.keys()
        assert is_sublist(
            ["x", "y", "width", "height"],
            coord.keys(),
        )

        return Box(
            id=id,
            class_name=class_name,
            class_id=self.object_classes_map[unique_string_builder(
                class_name,
                ClassType.BOX,
            )]["id"],
            coord=coord,
            meta=meta if meta else self._get_meta(class_name, ClassType.BOX),
            properties_def=self.object_classes_map[unique_string_builder(
                class_name,
                ClassType.BOX,
            )]["properties"],
            properties=properties,
            tracking_id=tracking_id,
        )

    def from_dict(self, box):
        class_name = box["class_name"]
        assert unique_string_builder(
            class_name,
            ClassType.BOX,
        ) in self.object_classes_map.keys()
        assert is_sublist(
            ["x", "y", "width", "height"],
            box["annotation"]["coord"].keys(),
        )

        properties = box["properties"] if "properties" in box else []
        return Box(
            id=box["id"] if "id" in box else str(uuid4()),
            class_name=class_name,
            class_id=self.object_classes_map[unique_string_builder(
                class_name,
                ClassType.BOX,
            )]["id"],
            coord=box["annotation"]["coord"],
            meta=box["annotation"]["meta"] if "meta" in box["annotation"] else self._get_meta(
                class_name,
                ClassType.BOX,
            ),
            properties_def=self.object_classes_map[unique_string_builder(
                class_name,
                ClassType.BOX,
            )]["properties"],
            properties=properties,
            tracking_id=box["tracking_id"] if "tracking_id" in box else 0,
        )


class RotatedBoxCreator(LabelCreator):
    def __init__(self, object_classes_map):
        super().__init__(object_classes_map)

    def get_default_rotated_box_coord(self):
        return {
            "cx": 0.0,
            "cy": 0.0,
            "width": 0.0,
            "height": 0.0,
            "angle": 0.0,
        }

    def create(
        self,
        class_name: str,
        coord: dict,
        meta: dict = {},
        properties: list = [],
        id: str = str(uuid4()),
        tracking_id: int = None,
    ):
        assert unique_string_builder(
            class_name,
            ClassType.RBOX,
        ) in self.object_classes_map.keys()
        assert is_sublist(
            ["cx", "cy", "width", "height", "angle"],
            coord.keys(),
        )

        return RotatedBox(
            id=id,
            class_name=class_name,
            class_id=self.object_classes_map[unique_string_builder(
                class_name,
                ClassType.RBOX,
            )]["id"],
            coord=coord,
            meta=meta if meta else self._get_meta(class_name, ClassType.RBOX),
            properties_def=self.object_classes_map[unique_string_builder(
                class_name,
                ClassType.RBOX,
            )]["properties"],
            properties=properties,
            tracking_id=tracking_id,
        )

    def from_dict(self, rbox):
        class_name = rbox["class_name"]
        assert unique_string_builder(
            class_name,
            ClassType.RBOX,
        ) in self.object_classes_map.keys()
        assert is_sublist(
            ["cx", "cy", "width", "height", "angle"],
            rbox["annotation"]["coord"].keys()
        )

        properties = rbox["properties"] if "properties" in rbox else []
        return RotatedBox(
            id=rbox["id"] if "id" in rbox else str(uuid4()),
            class_name=class_name,
            class_id=self.object_classes_map[unique_string_builder(
                class_name,
                ClassType.RBOX,
            )]["id"],
            coord=rbox["annotation"]["coord"],
            meta=rbox["annotation"]["meta"]
            if "meta" in rbox["annotation"]
            else self._get_meta(class_name, ClassType.RBOX),
            properties_def=self.object_classes_map[unique_string_builder(
                class_name,
                ClassType.RBOX,
            )]["properties"],
            properties=properties,
            tracking_id=rbox["tracking_id"] if "tracking_id" in rbox else 0,
        )


class PolylineCreator(LabelCreator):
    def __init__(self, object_classes_map):
        super().__init__(object_classes_map)

    def get_default_polyline_coord(self):
        return {"points": [[{"x": 0, "y": 0}]]}

    def create(
        self,
        class_name: str,
        coord: dict,
        meta: dict = {},
        properties: list = [],
        id: str = str(uuid4()),
        tracking_id: int = None,
    ):
        assert unique_string_builder(
            class_name,
            ClassType.POLYLINE,
        ) in self.object_classes_map.keys()
        assert is_sublist(
            ["points"],
            coord.keys(),
        )
        assert len(dim(coord["points"])) == 2
        assert dim(coord["points"])[-1] >= 1

        return Polyline(
            id=id,
            class_name=class_name,
            class_id=self.object_classes_map[unique_string_builder(
                class_name,
                ClassType.POLYLINE,
            )]["id"],
            coord=coord,
            multiple=True,
            meta=meta if meta else self._get_meta(
                class_name, ClassType.POLYLINE),
            properties_def=self.object_classes_map[unique_string_builder(
                class_name,
                ClassType.POLYLINE,
            )]["properties"],
            properties=properties,
            tracking_id=tracking_id,
        )

    def from_dict(self, polyline):
        class_name = polyline["class_name"]
        assert unique_string_builder(
            class_name,
            ClassType.POLYLINE,
        ) in self.object_classes_map.keys()
        assert is_sublist(
            ["points"],
            polyline["annotation"]["coord"].keys(),
        )
        assert len(dim(polyline["annotation"]["coord"]["points"])) == 2
        assert dim(polyline["annotation"]["coord"]["points"])[-1] >= 1
        properties = polyline["properties"] if "properties" in polyline else []

        return Polyline(
            id=polyline["id"] if "id" in polyline else str(uuid4()),
            class_name=class_name,
            class_id=self.object_classes_map[unique_string_builder(
                class_name,
                ClassType.POLYLINE
            )]["id"],
            coord=polyline["annotation"]["coord"],
            multiple=True,
            meta=polyline["annotation"]["meta"]
            if "meta" in polyline["annotation"]
            else self._get_meta(class_name, ClassType.POLYLINE),
            properties_def=self.object_classes_map[unique_string_builder(
                class_name,
                ClassType.POLYLINE
            )]["properties"],
            properties=properties,
            tracking_id=polyline["tracking_id"]
            if "tracking_id" in polyline
            else 0,
        )


class PolygonCreator(LabelCreator):
    def __init__(self, object_classes_map):
        super().__init__(object_classes_map)

    def get_default_polyline_coord(self):
        return {"points": [[[{"x": 0, "y": 0}]]]}

    def create(
        self,
        class_name: str,
        coord: dict,
        meta: dict = {},
        properties: list = [],
        id: str = str(uuid4()),
        tracking_id: int = None,
    ):
        assert unique_string_builder(
            class_name,
            ClassType.POLYGON,
        ) in self.object_classes_map.keys()
        assert is_sublist(
            ["points"],
            coord.keys(),
        )
        assert len(dim(coord["points"])) == 3
        assert dim(coord["points"])[-1] >= 3

        return Polygon(
            id=id,
            class_name=class_name,
            class_id=self.object_classes_map[unique_string_builder(
                class_name,
                ClassType.POLYGON,
            )]["id"],
            coord=coord,
            multiple=True,
            meta=meta if meta else self._get_meta(
                class_name, ClassType.POLYGON),
            properties_def=self.object_classes_map[unique_string_builder(
                class_name,
                ClassType.POLYGON,
            )]["properties"],
            properties=properties,
            tracking_id=tracking_id,
        )

    def from_dict(self, polygon):
        class_name = polygon["class_name"]
        assert unique_string_builder(
            class_name,
            ClassType.POLYGON,
        ) in self.object_classes_map.keys()
        assert is_sublist(
            ["points"],
            polygon["annotation"]["coord"].keys(),
        )
        assert len(dim(polygon["annotation"]["coord"]["points"])) == 3
        assert dim(polygon["annotation"]["coord"]["points"])[-1] >= 3

        properties = polygon["properties"] if "properties" in polygon else []

        return Polygon(
            id=polygon["id"] if "id" in polygon else str(uuid4()),
            class_name=class_name,
            class_id=self.object_classes_map[unique_string_builder(
                class_name,
                ClassType.POLYGON,
            )]["id"],
            coord=polygon["annotation"]["coord"],
            multiple=True,
            meta=polygon["annotation"].get("meta", None)
            if "meta" in polygon["annotation"]
            else self._get_meta(class_name, ClassType.POLYGON),
            properties_def=self.object_classes_map[unique_string_builder(
                class_name,
                ClassType.POLYGON,
            )]["properties"],
            properties=properties,
            tracking_id=polygon["tracking_id"]
            if "tracking_id" in polygon
            else 0,
        )


class KeypointCreator(LabelCreator):
    def __init__(self, object_classes_map, keypoint_map):
        super().__init__(object_classes_map)
        self.keypoint_map = keypoint_map

    def get_default_keypoint_coord(self, keypoint_interface_id):
        return {
            "points": [
                {
                    "name": point["name"],
                    "x": 0.0,
                    "y": 0.0,
                    "state": {
                        "visible": True,
                        "valid": True,
                    },
                }
                for point in self.keypoint_map[keypoint_interface_id]["points"]
            ]
        }

    def get_default_keypoint_legacy_coord(self, keypoint_interface_id):
        return {
            "graph": {
                "nodes": [
                    {
                        "x": 0.0,
                        "y": 0.0,
                    }
                    for _ in self.keypoint_map[keypoint_interface_id]["points"]
                ],
                "states": [
                    {"state": {"visible": True}}
                    for _ in self.keypoint_map[keypoint_interface_id]["points"]
                ],
            }
        }

    def create(
        self,
        class_name: str,
        coord: dict,
        meta: dict = {},
        properties: list = [],
        id: str = str(uuid4()),
        tracking_id: int = None,
    ):
        assert unique_string_builder(
            class_name,
            ClassType.KEYPOINT,
        ) in self.object_classes_map.keys()
        assert is_sublist(
            ["points"],
            coord.keys(),
        )
        assert is_sublist(
            ["x", "y"],
            coord["points"][0].keys(),
        )

        return Keypoint(
            id=id,
            class_name=class_name,
            class_id=self.object_classes_map[unique_string_builder(
                class_name,
                ClassType.KEYPOINT,
            )]["id"],
            coord=coord,
            meta=meta if meta else self._get_meta(
                class_name, ClassType.KEYPOINT),
            properties_def=self.object_classes_map[unique_string_builder(
                class_name,
                ClassType.KEYPOINT,
            )]["properties"],
            properties=properties,
            tracking_id=tracking_id,
        )

    def from_dict(self, keypoint):
        class_name = keypoint["class_name"]
        coord = keypoint["annotation"]["coord"]

        assert unique_string_builder(
            class_name,
            ClassType.KEYPOINT,
        ) in self.object_classes_map.keys()
        assert is_sublist(
            ["points"],
            coord.keys(),
        )
        assert is_sublist(
            ["x", "y"],
            coord["points"][0].keys(),
        )

        return Keypoint(
            id=keypoint["id"] if "id" in keypoint else str(uuid4()),
            class_name=class_name,
            class_id=self.object_classes_map[unique_string_builder(
                class_name,
                ClassType.KEYPOINT,
            )]["id"],
            coord=coord,
            meta=keypoint["annotation"]["meta"]
            if "meta" in keypoint["annotation"]
            else self._get_meta(class_name, ClassType.KEYPOINT),
            properties_def=self.object_classes_map[unique_string_builder(
                class_name,
                ClassType.KEYPOINT,
            )]["properties"],
            properties=keypoint["properties"] if "properties" in keypoint else [
            ],
            tracking_id=keypoint["tracking_id"]
            if "tracking_id" in keypoint
            else 0,
        )


class Cuboid2DCreator(LabelCreator):
    def __init__(self, object_classes_map):
        super().__init__(object_classes_map)

    def get_default_cuboid2d_coord(self):
        return {
            "near": {
                "x": 0.0,
                "y": 0.0,
                "width": 0.0,
                "height": 0.0,
            },
            "far": {
                "x": 0.0,
                "y": 0.0,
                "width": 0.0,
                "height": 0.0,
            },
        }

    def create(
        self,
        class_name: str,
        coord: dict,
        meta: dict = {},
        properties: list = [],
        id: str = str(uuid4()),
        tracking_id: int = None,
    ):
        assert unique_string_builder(
            class_name,
            ClassType.CUBOID2D,
        ) in self.object_classes_map.keys()
        assert is_sublist(
            ["near", "far"],
            coord.keys(),
        )

        return Cuboid2D(
            id=id,
            class_name=class_name,
            class_id=self.object_classes_map[unique_string_builder(
                class_name,
                ClassType.CUBOID2D,
            )]["id"],
            coord=coord,
            meta=meta if meta else self._get_meta(
                class_name,
                ClassType.CUBOID2D
            ),
            properties_def=self.object_classes_map[unique_string_builder(
                class_name,
                ClassType.CUBOID2D,
            )]["properties"],
            properties=properties,
            tracking_id=tracking_id,
        )

    def from_dict(self, cuboid2d):
        class_name = cuboid2d["class_name"]
        assert unique_string_builder(
            class_name,
            ClassType.CUBOID2D,
        ) in self.object_classes_map.keys()
        assert is_sublist(
            ["near", "far"],
            cuboid2d["annotation"]["coord"].keys()
        )

        properties = cuboid2d["properties"] if "properties" in cuboid2d else []
        a = Cuboid2D(
            id=cuboid2d["id"] if "id" in cuboid2d else str(uuid4()),
            class_name=class_name,
            class_id=self.object_classes_map[unique_string_builder(
                class_name,
                ClassType.CUBOID2D,
            )]["id"],
            coord=cuboid2d["annotation"]["coord"],
            meta=cuboid2d["annotation"]["meta"]
            if "meta" in cuboid2d["annotation"]
            else self._get_meta(class_name, ClassType.CUBOID2D),
            properties_def=self.object_classes_map[unique_string_builder(
                class_name,
                ClassType.CUBOID2D,
            )]["properties"],
            properties=properties,
            tracking_id=cuboid2d["tracking_id"]
            if "tracking_id" in cuboid2d
            else 0,
        )
        return a


class AutoLabelCreator(LabelCreator):
    def __init__(self, object_classes_map, keypoint_map):
        super().__init__(object_classes_map)
        self.keypoint_map = keypoint_map

    def create(self):
        raise NotImplementedError

    def from_dict(self, label):
        creators = [
            BoxCreator(self.object_classes_map),
            RotatedBoxCreator(self.object_classes_map),
            PolylineCreator(self.object_classes_map),
            PolygonCreator(self.object_classes_map),
            KeypointCreator(self.object_classes_map, self.keypoint_map),
            Cuboid2DCreator(self.object_classes_map),
        ]
        for creator in creators:
            try:
                return creator.from_dict(label)
            except:
                continue
        return None



from phy_credit.common.utils import (
    expand_options,
    remove_free_response,
)
from phy_credit.common.video_utils import (
    calculate_video_properties_count,
    flatten_video_obj_properties,
    preprocess_video_obj_properties,
)


# preprocessing: step 1
def test_flatten_video_object_properties(mock_video_objects, num_properties_flattened):
    assert len(flatten_video_obj_properties([])) == 0
    assert (
        len(flatten_video_obj_properties(mock_video_objects))
        == num_properties_flattened
    )


# preprocessing: step 2
def test_remove_free_response(mock_video_objects, num_properties_removed_free_response):
    input = flatten_video_obj_properties(mock_video_objects)
    assert len(remove_free_response(input)
               ) == num_properties_removed_free_response


# preprocessing: step 3
def test_expand_options(mock_video_objects, num_properties_expanded_options):
    input = remove_free_response(
        flatten_video_obj_properties(mock_video_objects))
    assert len(expand_options(input)) == num_properties_expanded_options


# preprocessing: combined steps
def test_preprocess_video_objects(mock_video_objects, num_properties_preprocessed):
    result = preprocess_video_obj_properties(mock_video_objects)
    assert len(result) == num_properties_preprocessed


# counting: combined steps
def test_calculate_video_properties_count(mock_video_classes_properties_count, mock_video_objects):
    assert calculate_video_properties_count([]) == []
    expected = mock_video_classes_properties_count
    for (result, expected) in zip(
        calculate_video_properties_count(mock_video_objects), expected
    ):
        assert result == expected



from typing import Any, Dict, List
from ..common.utils import (add_class_id, count_properties,
                            expand_options, remove_free_response)

JsonDict = Dict[str, Any]


def flatten_img_obj_properties(objects: List[JsonDict]) -> List[JsonDict]:
    return [
        add_class_id(property, obj)
        for obj in objects
        for property in obj["properties"]
    ]


def preprocess_image_obj_properties(objects):
    return expand_options(remove_free_response(flatten_img_obj_properties(objects)))


def calculate_imageV2_properties_count(objects):
    properties = preprocess_image_obj_properties(objects)
    return count_properties(properties)



from phy_credit.common.image_utils import (
    calculate_imageV2_properties_count,
    count_properties,
)


def test_count_properties(mock_preprocessed_properties, mock_count_properties_result):
    input = mock_preprocessed_properties
    expected = mock_count_properties_result
    assert all([a == b for a, b in zip(count_properties(input), expected)])


def test_calculate_image_properties_count(mock_image_objects_with_properties, mock_image_classes_properties_count):
    objects = mock_image_objects_with_properties
    expected = mock_image_classes_properties_count
    result = calculate_imageV2_properties_count(objects)
    assert all([a == b for a, b in zip(result, expected)])



import re, ast
from setuptools import setup, find_packages


def version(name):
    _re = re.compile(r'__version__\s+=\s+(.*)')
    with open(f'{name}/__init__.py', 'rb') as f:
        return str(ast.literal_eval(_re.search(f.read().decode('utf-8')).group(1)))


setup(
    name='phy_credit',
    version=version('phy_credit'),
    author="Superb AI",
    packages=find_packages(),
    install_requires=[
        'semver>=2.13.0',
    ],
)
